import base64
import os
import random
import string
import uuid
import xml.etree.ElementTree as ET
from datetime import datetime

import requests
from requests.sessions import Request

from uup.composer import (CallerAttributesComposer, DeviceAttributesComposer,
                          ProductsComposer)
from uup.exception import CookieNotFoundException


class UnifiedUpdatePlatformAPI:

    def __init__(self):
        self.cookie = None
        self.protocol_version = '2.50'
        self.xml_namespaces = {
            'a': 'http://www.w3.org/2005/08/addressing',
            's': 'http://www.w3.org/2003/05/soap-envelope',
            'o': 'http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-secext-1.0.xsd',
            'wsu': 'http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-utility-1.0.xsd',
            'wuws': 'http://schemas.microsoft.com/msus/2014/10/WindowsUpdateAuthorization',
            'cws': 'http://www.microsoft.com/SoftwareDistribution/Server/ClientWebService'
        }

    def _compose_xml(self, action: str, *, secured: bool = False) -> ET.Element:
        xml_root = ET.parse(os.path.dirname(os.path.realpath(__file__)) + '/template/base.xml').getroot()

        xml_header = xml_root.find('./s:Header', self.xml_namespaces)
        xml_header.find('./a:Action', self.xml_namespaces).text = action
        xml_header.find('./a:MessageID', self.xml_namespaces).text = str(uuid.uuid4())
        xml_header.find('./a:To', self.xml_namespaces).text = 'https://fe3cr.delivery.mp.microsoft.com/ClientWebService/client.asmx' + ('/secured' if secured else '')
        xml_header.find('./o:Security/wsu:Timestamp/wsu:Created', self.xml_namespaces).text = datetime.fromtimestamp(datetime.utcnow().timestamp()).strftime('%Y-%m-%dT%H:%M:%SZ')
        xml_header.find('./o:Security/wsu:Timestamp/wsu:Expires', self.xml_namespaces).text = datetime.fromtimestamp(datetime.utcnow().timestamp() + 300).strftime('%Y-%m-%dT%H:%M:%SZ')
        xml_header.find('./o:Security/wuws:WindowsUpdateTicketsToken/TicketType/Device', self.xml_namespaces).text = self._generate_device()

        return xml_root

    def _generate_device(self) -> str:
        device_start = '13003002c377040014d5bcac7a66de0d50beddf9bba16c87edb9e019898000'
        device_random = ''.join(random.choices(string.hexdigits, k=1054))
        device_end = 'b401'

        device = base64.b64encode(bytes.fromhex(device_start + device_random + device_end)).decode('utf-8')
        device = base64.b64encode(f't={device}&p='.encode()).decode('utf-8')

        return device

    def _send_request(self, payload: str) -> Request:
        return requests.post('https://fe3cr.delivery.mp.microsoft.com/ClientWebService/client.asmx', data=payload, headers={'Content-Type': 'application/soap+xml; charset=utf-8', 'User-Agent': f'Windows-Update-Agent/10.0.10011.16384 Client-Protocol/{self.protocol_version}'}, verify=False)

    def get_cookie(self) -> Request:
        xml_root = self._compose_xml('http://www.microsoft.com/SoftwareDistribution/Server/ClientWebService/GetCookie')

        xml_root.find('./s:Body', self.xml_namespaces).append(ET.parse(os.path.dirname(os.path.realpath(__file__)) + '/template/get_cookie.xml').getroot())

        xml_get_cookie = xml_root.find('./s:Body/cws:GetCookie', self.xml_namespaces)
        xml_get_cookie.find('./cws:oldCookie/cws:Expiration', self.xml_namespaces).text = datetime.fromtimestamp(datetime.utcnow().timestamp()).strftime('%Y-%m-%dT%H:%M:%SZ')
        xml_get_cookie.find('./cws:lastChange', self.xml_namespaces).text = datetime.fromtimestamp(datetime.utcnow().timestamp()).strftime('%Y-%m-%dT%H:%M:%SZ')
        xml_get_cookie.find('./cws:currentTime', self.xml_namespaces).text = datetime.fromtimestamp(datetime.utcnow().timestamp()).strftime('%Y-%m-%dT%H:%M:%SZ')
        xml_get_cookie.find('./cws:protocolVersion', self.xml_namespaces).text = self.protocol_version

        return self._send_request(ET.tostring(xml_root, encoding='utf-8'))

    def get_extended_update_info2(self, update_id: str, revision_number: int, device_attributes: str) -> Request:
        xml_root = self._compose_xml('http://www.microsoft.com/SoftwareDistribution/Server/ClientWebService/GetExtendedUpdateInfo2', secured=True)

        xml_root.find('./s:Body', self.xml_namespaces).append(ET.parse(os.path.dirname(os.path.realpath(__file__)) + '/template/get_extended_update_info2.xml').getroot())

        xml_get_extended_update_info2 = xml_root.find('./s:Body/cws:GetExtendedUpdateInfo2', self.xml_namespaces)
        xml_get_extended_update_info2.find('./cws:updateIDs/cws:UpdateIdentity/cws:UpdateID', self.xml_namespaces).text = update_id
        xml_get_extended_update_info2.find('./cws:updateIDs/cws:UpdateIdentity/cws:RevisionNumber', self.xml_namespaces).text = str(revision_number)

        xml_info_types = xml_get_extended_update_info2.find('./cws:infoTypes', self.xml_namespaces)
        for xml_update_fragment_type in ['FileUrl', 'FileDescription', 'EsrpDecryptionInformation', 'PiecesHashUrl', 'BlockMapUrl']:
            ET.SubElement(xml_info_types, '{http://www.microsoft.com/SoftwareDistribution/Server/ClientWebService}XmlUpdateFragmentType').text = xml_update_fragment_type

        xml_get_extended_update_info2.find('./cws:deviceAttributes', self.xml_namespaces).text = device_attributes

        return self._send_request(ET.tostring(xml_root, encoding='utf-8'))

    def set_cookie(self, encrypted_data: str, expiration: str) -> None:
        self.cookie = {'encrypted_data': encrypted_data, 'expiration': expiration}

    def sync_updates(self, build: str, sku_id: int, arches: list, channel: str, *, defer_updates: int = 0) -> Request:
        if not self.cookie:
            raise CookieNotFoundException()

        xml_root = self._compose_xml('http://www.microsoft.com/SoftwareDistribution/Server/ClientWebService/SyncUpdates')

        xml_root.find('./s:Body', self.xml_namespaces).append(ET.parse(os.path.dirname(os.path.realpath(__file__)) + '/template/sync_updates.xml').getroot())

        xml_sync_updates = xml_root.find('./s:Body/cws:SyncUpdates', self.xml_namespaces)
        xml_sync_updates.find('./cws:cookie/cws:Expiration', self.xml_namespaces).text = self.cookie['expiration']
        xml_sync_updates.find('./cws:cookie/cws:EncryptedData', self.xml_namespaces).text = self.cookie['encrypted_data']
        xml_sync_updates.find('./cws:parameters/cws:ExpressQuery', self.xml_namespaces).text = 'false'
        xml_sync_updates.find('./cws:parameters/cws:SkipSoftwareSync', self.xml_namespaces).text = 'false'
        xml_sync_updates.find('./cws:parameters/cws:NeedTwoGroupOutOfScopeUpdates', self.xml_namespaces).text = 'true'
        xml_sync_updates.find('./cws:parameters/cws:AlsoPerformRegularSync', self.xml_namespaces).text = 'true'

        xml_xml_update_fragment_types = xml_sync_updates.find('./cws:parameters/cws:ExtendedUpdateInfoParameters/cws:XmlUpdateFragmentTypes', self.xml_namespaces)
        for xml_update_fragment_type in ['Extended', 'LocalizedProperties']:
            ET.SubElement(xml_xml_update_fragment_types, '{http://www.microsoft.com/SoftwareDistribution/Server/ClientWebService}XmlUpdateFragmentType').text = xml_update_fragment_type

        xml_locales = xml_sync_updates.find('./cws:parameters/cws:ExtendedUpdateInfoParameters/cws:Locales', self.xml_namespaces)
        for locale in ['en-US']:
            ET.SubElement(xml_locales, '{http://www.microsoft.com/SoftwareDistribution/Server/ClientWebService}string').text = locale

        xml_sync_updates.find('./cws:parameters/cws:ProductsParameters/cws:SyncCurrentVersionOnly', self.xml_namespaces).text = 'false'

        caller_attributes_composer = CallerAttributesComposer()
        xml_sync_updates.find('./cws:parameters/cws:ProductsParameters/cws:CallerAttributes', self.xml_namespaces).text = caller_attributes_composer.compose()

        device_attributes_composer = DeviceAttributesComposer()
        xml_sync_updates.find('./cws:parameters/cws:ProductsParameters/cws:DeviceAttributes', self.xml_namespaces).text = device_attributes_composer.compose(build, sku_id, channel, defer_updates)

        products_composer = ProductsComposer()
        xml_sync_updates.find('./cws:parameters/cws:ProductsParameters/cws:Products', self.xml_namespaces).text = products_composer.compose(build, sku_id, arches)

        return self._send_request(ET.tostring(xml_root, encoding='utf-8'))
